"""
OONI data format.

See:

- https://github.com/ooni/spec/tree/master/data-formats

- https://github.com/ooni/spec/tree/master/nettests
"""
import logging
import hashlib

from pathlib import Path
from base64 import b64decode

from datetime import datetime
from typing import Optional, Tuple, Union, List, Union, Dict

from dataclasses import dataclass

import orjson
from mashumaro.config import BaseConfig, TO_DICT_ADD_OMIT_NONE_FLAG
from mashumaro import DataClassDictMixin


log = logging.getLogger("oonidata.dataformat")


class BaseModel(DataClassDictMixin):
    class Config(BaseConfig):
        # This makes it possible to call .to_dict(omit_none=True) to remove any
        # attributes of the dataclass that a None, saving up quite a bit of
        # space for unnecessary keys
        code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]


@dataclass
class BinaryData(BaseModel):
    format: str
    data: str


MaybeBinaryData = Union[str, BinaryData, None]
Failure = Optional[str]


def guess_decode(s: bytes) -> str:
    """
    best effort decoding of a string of bytes
    """
    for encoding in ("ascii", "utf-8", "latin1"):
        try:
            return s.decode(encoding)
        except UnicodeDecodeError:
            pass
    log.warning(f"unable to decode '{s}'")
    return s.decode("ascii", "ignore")


def maybe_binary_data_to_str(mbd: Union[MaybeBinaryData, Dict]) -> str:
    if isinstance(mbd, BinaryData):
        return guess_decode(b64decode(mbd.data))
    elif isinstance(mbd, dict):
        return guess_decode(b64decode(mbd["data"]))
    elif isinstance(mbd, str):
        return mbd

    raise Exception(f"Invalid type {type(mbd)} {mbd}")


def maybe_binary_data_to_bytes(mbd: Union[MaybeBinaryData, Dict]) -> bytes:
    if isinstance(mbd, BinaryData):
        return b64decode(mbd.data)
    elif isinstance(mbd, dict):
        return b64decode(mbd["data"])
    elif isinstance(mbd, str):
        return mbd.encode("utf-8")

    raise Exception(f"Invalid type {type(mbd)} {mbd}")


def trivial_id(raw: bytes, msm: Dict) -> str:
    """Generate a trivial id of the measurement to allow upsert if needed
    This is used for legacy (before measurement_uid) measurements
    - Deterministic / stateless with no DB interaction
    - Malicious/bugged msmts with collisions on report_id/input/test_name lead
    to different hash values avoiding the collision
    - Malicious/duplicated msmts that are semantically identical to the "real"
    one lead to harmless collisions
    - Sortable by date
    """
    VER = "01"
    h = hashlib.shake_128(raw).hexdigest(15)
    try:
        t = msm.get("measurement_start_time") or ""
        t = datetime.strptime(t, "%Y-%m-%d %H:%M:%S")
        ts = t.strftime("%Y%m%d")
    except:
        ts = "00000000"
    tid = f"{VER}{ts}{h}"
    return tid


@dataclass
class BaseTestKeys(BaseModel):
    client_resolver: Optional[str] = None


@dataclass
class BaseMeasurement(BaseModel):
    annotations: Dict[str, str]

    input: Union[str, List[str], None]
    report_id: str

    measurement_start_time: str
    test_start_time: str

    probe_asn: str
    probe_cc: str
    probe_ip: Optional[str]

    test_name: str
    test_version: str
    test_runtime: float

    software_name: str
    software_version: str

    test_keys: BaseTestKeys

    resolver_asn: Optional[str] = None
    resolver_ip: Optional[str] = None
    resolver_network_name: Optional[str] = None

    probe_network_name: Optional[str] = None

    test_helpers: Optional[Dict] = None
    data_format_version: Optional[str] = None
    measurement_uid: Optional[str] = None


# This is not 100% accurate, ideally we would say
# List[Tuple[str, MaybeBinaryData]], yet this doesn't work because we don't have
# tuples in JSON
HeadersList = List[List[Union[str, MaybeBinaryData]]]
HeadersListBytes = List[Tuple[str, bytes]]
HeadersListStr = List[Tuple[str, str]]


@dataclass
class TorInfo(BaseModel):
    is_tor: bool
    exit_ip: Optional[str]
    exit_name: Optional[str]


@dataclass
class HTTPBase(BaseModel):
    body: MaybeBinaryData = None
    body_is_truncated: Optional[bool] = None
    headers: Optional[Dict[str, str]] = None
    headers_list: Optional[HeadersList] = None

    _body_bytes = None
    _body_str = None
    _headers = None
    _headers_list_bytes = None
    _headers_list_str = None

    @property
    def body_str(self) -> Optional[str]:
        if not self.body:
            return None

        if self._body_str:
            return self._body_str

        self._body_str = maybe_binary_data_to_str(self.body)
        return self._body_str

    @property
    def body_bytes(self) -> Optional[bytes]:
        if not self.body:
            return None

        if self._body_bytes:
            return self._body_bytes

        self._body_bytes = maybe_binary_data_to_bytes(self.body)
        return self._body_bytes

    @property
    def headers_str(self) -> Optional[Dict[str, str]]:
        if not self.headers_list_str:
            return None
        return {k: v for k, v in self.headers_list_str}

    @property
    def headers_bytes(self) -> Optional[Dict[str, bytes]]:
        if not self.headers_list_bytes:
            return None
        return {k: v for k, v in self.headers_list_bytes}

    @property
    def headers_list_str(self) -> Optional[List[Tuple[str, str]]]:
        if not self.headers_list:
            return None

        if self._headers_list_str:
            return self._headers_list_str

        self._headers_list_str = [
            (maybe_binary_data_to_str(k), maybe_binary_data_to_str(v))
            for k, v in self.headers_list
        ]
        return self._headers_list_str

    @property
    def headers_list_bytes(self) -> Optional[List[Tuple[str, bytes]]]:
        if not self.headers_list:
            return None

        if self._headers_list_bytes:
            return self._headers_list_bytes

        self._headers_list_bytes = [
            (guess_decode(maybe_binary_data_to_bytes(k)), maybe_binary_data_to_bytes(v))
            for k, v in self.headers_list
        ]
        return self._headers_list_bytes

    def __post_init__(self):
        if not self.headers_list and self.headers:
            self.headers_list = []
            for k, v in self.headers.items():
                self.headers_list.append([k, v])


@dataclass
class HTTPRequest(HTTPBase):
    url: str = ""
    method: Optional[str] = None
    tor: Optional[TorInfo] = None
    x_transport: Optional[str] = "tcp"


@dataclass
class HTTPResponse(HTTPBase):
    code: Optional[int] = None


@dataclass
class HTTPTransaction(BaseModel):
    failure: Failure = None

    request: Optional[HTTPRequest] = None
    response: Optional[HTTPResponse] = None

    network: Optional[str] = None
    address: Optional[str] = None
    alpn: Optional[str] = None

    transaction_id: Optional[int] = None

    t: Optional[float] = None
    t0: Optional[float] = None

    def response_sha1(self) -> str:
        if self.response and self.response.body_bytes:
            return hashlib.sha1(self.response.body_bytes).hexdigest()
        return ""


@dataclass
class DNSAnswer(BaseModel):
    answer_type: str
    asn: Optional[int] = None
    as_org_name: Optional[str] = None
    hostname: Optional[str] = None
    ipv4: Optional[str] = None
    ipv6: Optional[str] = None
    ttl: Optional[int] = None

    # Deprecated
    expiration_limit: Optional[str] = None
    minimum_ttl: Optional[str] = None
    refresh_interval: Optional[str] = None
    responsible_name: Optional[str] = None
    retry_interval: Optional[str] = None
    serial_number: Optional[str] = None


@dataclass
class DNSQuery(BaseModel):
    hostname: str
    query_type: str

    failure: Failure = None
    engine: Optional[str] = None

    answers: Optional[List[DNSAnswer]] = None

    raw_response: Optional[str] = None
    rcode: Optional[int] = None

    resolver_address: Optional[str] = None

    transaction_id: Optional[int] = None

    t: Optional[float] = None
    t0: Optional[float] = None

    # Deprecated field
    dial_id: Optional[int] = None


@dataclass
class TCPConnectStatus(BaseModel):
    success: bool
    blocked: Optional[bool] = None
    failure: Failure = None


@dataclass
class TCPConnect(BaseModel):
    ip: str
    port: int
    status: TCPConnectStatus

    transaction_id: Optional[int] = None

    t: Optional[float] = None
    t0: Optional[float] = None

    # Deprecated fields
    conn_id: Optional[int] = None
    dial_id: Optional[int] = None


@dataclass
class TLSHandshake(BaseModel):
    network: Optional[str] = None
    address: Optional[str] = None
    cipher_suite: Optional[str] = None

    failure: Failure = None
    so_error: Failure = None

    negotiated_protocol: Optional[str] = None

    no_tls_verify: Optional[bool] = None
    peer_certificates: Optional[List[BinaryData]] = None

    server_name: Optional[str] = None

    tags: Optional[List[str]] = None
    tls_version: Optional[str] = None

    t: Optional[float] = None
    t0: Optional[float] = None

    transaction_id: Optional[int] = None

    # Deprecated
    conn_id: Optional[int] = None


@dataclass
class NetworkEvent(BaseModel):
    operation: str
    t: float
    failure: Failure = None
    address: Optional[str] = None
    num_bytes: Optional[int] = None
    proto: Optional[str] = None
    tags: Optional[List[str]] = None
    transaction_id: Optional[str] = None

    # Deprecated fields
    dial_id: Optional[int] = None
    conn_id: Optional[int] = None


@dataclass
class WebConnectivityControlHTTPRequest(BaseModel):
    body_length: Optional[int] = None
    failure: Failure = None
    title: Optional[str] = None
    headers: Optional[Dict[str, str]] = None
    status_code: Optional[int] = None


@dataclass
class WebConnectivityControlDNS(BaseModel):
    failure: Failure = None
    addrs: Optional[List[str]] = None


@dataclass
class WebConnectivityControlTCPConnectStatus(BaseModel):
    status: Optional[bool] = None
    failure: Failure = None


@dataclass
class WebConnectivityControl(BaseModel):
    tcp_connect: Optional[Dict[str, WebConnectivityControlTCPConnectStatus]] = None
    http_request: Optional[WebConnectivityControlHTTPRequest] = None
    dns: Optional[WebConnectivityControlDNS] = None


@dataclass
class WebConnectivityTestKeys(BaseModel):
    dns_experiment_failure: Failure = None
    control_failure: Failure = None
    http_experiment_failure: Failure = None

    dns_consistency: Optional[str] = None

    body_length_match: Optional[bool] = None
    body_proportion: Optional[float] = None
    status_code_match: Optional[bool] = None
    headers_match: Optional[bool] = None
    title_match: Optional[bool] = None
    accessible: Optional[bool] = None
    blocking: Union[str, bool, None] = None

    control: Optional[WebConnectivityControl] = None
    tls_handshakes: Optional[List[TLSHandshake]] = None
    network_events: Optional[List[NetworkEvent]] = None
    queries: Optional[List[DNSQuery]] = None
    tcp_connect: Optional[List[TCPConnect]] = None
    requests: Optional[List[HTTPTransaction]] = None

    x_status: Optional[int] = None
    x_dns_runtime: Optional[int] = None
    x_th_runtime: Optional[int] = None
    x_tcptls_runtime: Optional[int] = None
    x_http_runtime: Optional[int] = None

    client_resolver: Optional[str] = None

    retries: Optional[int] = None
    socksproxy: Optional[str] = None


@dataclass
class WebConnectivity(BaseMeasurement):
    test_keys: WebConnectivityTestKeys


@dataclass
class WhatsappTestKeys(BaseTestKeys):
    failure: Optional[str] = None
    failed_operation: Optional[str] = None

    network_events: Optional[List[NetworkEvent]] = None
    tls_handshakes: Optional[List[TLSHandshake]] = None
    queries: Optional[List[DNSQuery]] = None
    tcp_connect: Optional[List[TCPConnect]] = None
    requests: Optional[List[HTTPTransaction]] = None

    registration_server_failure: Optional[str] = None
    registration_server_status: Optional[str] = None
    whatsapp_endpoints_status: Optional[str] = None
    whatsapp_endpoints_blocked: Optional[List[str]] = None
    whatsapp_endpoints_dns_inconsistent: Optional[List[str]] = None

    whatsapp_web_failure: Optional[str] = None
    whatsapp_web_status: Optional[str] = None


@dataclass
class Whatsapp(BaseMeasurement):
    test_keys: WhatsappTestKeys


SIGNAL_ROOT_CA_OLD = """-----BEGIN CERTIFICATE-----
MIID7zCCAtegAwIBAgIJAIm6LatK5PNiMA0GCSqGSIb3DQEBBQUAMIGNMQswCQYD
VQQGEwJVUzETMBEGA1UECAwKQ2FsaWZvcm5pYTEWMBQGA1UEBwwNU2FuIEZyYW5j
aXNjbzEdMBsGA1UECgwUT3BlbiBXaGlzcGVyIFN5c3RlbXMxHTAbBgNVBAsMFE9w
ZW4gV2hpc3BlciBTeXN0ZW1zMRMwEQYDVQQDDApUZXh0U2VjdXJlMB4XDTEzMDMy
NTIyMTgzNVoXDTIzMDMyMzIyMTgzNVowgY0xCzAJBgNVBAYTAlVTMRMwEQYDVQQI
DApDYWxpZm9ybmlhMRYwFAYDVQQHDA1TYW4gRnJhbmNpc2NvMR0wGwYDVQQKDBRP
cGVuIFdoaXNwZXIgU3lzdGVtczEdMBsGA1UECwwUT3BlbiBXaGlzcGVyIFN5c3Rl
bXMxEzARBgNVBAMMClRleHRTZWN1cmUwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
ggEKAoIBAQDBSWBpOCBDF0i4q2d4jAXkSXUGpbeWugVPQCjaL6qD9QDOxeW1afvf
Po863i6Crq1KDxHpB36EwzVcjwLkFTIMeo7t9s1FQolAt3mErV2U0vie6Ves+yj6
grSfxwIDAcdsKmI0a1SQCZlr3Q1tcHAkAKFRxYNawADyps5B+Zmqcgf653TXS5/0
IPPQLocLn8GWLwOYNnYfBvILKDMItmZTtEbucdigxEA9mfIvvHADEbteLtVgwBm9
R5vVvtwrD6CCxI3pgH7EH7kMP0Od93wLisvn1yhHY7FuYlrkYqdkMvWUrKoASVw4
jb69vaeJCUdU+HCoXOSP1PQcL6WenNCHAgMBAAGjUDBOMB0GA1UdDgQWBBQBixjx
P/s5GURuhYa+lGUypzI8kDAfBgNVHSMEGDAWgBQBixjxP/s5GURuhYa+lGUypzI8
kDAMBgNVHRMEBTADAQH/MA0GCSqGSIb3DQEBBQUAA4IBAQB+Hr4hC56m0LvJAu1R
K6NuPDbTMEN7/jMojFHxH4P3XPFfupjR+bkDq0pPOU6JjIxnrD1XD/EVmTTaTVY5
iOheyv7UzJOefb2pLOc9qsuvI4fnaESh9bhzln+LXxtCrRPGhkxA1IMIo3J/s2WF
/KVYZyciu6b4ubJ91XPAuBNZwImug7/srWvbpk0hq6A6z140WTVSKtJG7EP41kJe
/oF4usY5J7LPkxK3LWzMJnb5EIJDmRvyH8pyRwWg6Qm6qiGFaI4nL8QU4La1x2en
4DGXRaLMPRwjELNgQPodR38zoCMuA8gHZfZYYoZ7D7Q1wNUiVHcxuFrEeBaYJbLE
rwLV
-----END CERTIFICATE-----""".encode(
    "ascii"
)

SIGNAL_ROOT_CA_NEW = """-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----""".encode(
    "ascii"
)

SIGNAL_PEM_STORE = (SIGNAL_ROOT_CA_OLD, SIGNAL_ROOT_CA_NEW)


@dataclass
class TCPTTestKeys(BaseTestKeys):
    received: Optional[List[MaybeBinaryData]] = None
    sent: Optional[List[MaybeBinaryData]] = None


@dataclass
class SignalTestKeys(BaseTestKeys):
    failure: Optional[str] = None
    failed_operation: Optional[str] = None

    network_events: Optional[List[NetworkEvent]] = None
    tls_handshakes: Optional[List[TLSHandshake]] = None
    queries: Optional[List[DNSQuery]] = None
    tcp_connect: Optional[List[TCPConnect]] = None
    requests: Optional[List[HTTPTransaction]] = None

    signal_backend_status: Optional[str] = None
    signal_backend_failure: Optional[str] = None


@dataclass
class Signal(BaseMeasurement):
    test_keys: SignalTestKeys


@dataclass
class URLGetterTestKeys(BaseTestKeys):
    failure: Failure = None
    socksproxy: Optional[str] = None
    tls_handshakes: Optional[List[TLSHandshake]] = None
    network_events: Optional[List[NetworkEvent]] = None
    queries: Optional[List[DNSQuery]] = None
    tcp_connect: Optional[List[TCPConnect]] = None
    requests: Optional[List[HTTPTransaction]] = None


@dataclass
class DNSCheckTestKeys(BaseTestKeys):
    lookups: Optional[Dict[str, URLGetterTestKeys]] = None
    bootstrap: Optional[URLGetterTestKeys] = None
    bootstrap_failure: Optional[str] = None


@dataclass
class DNSCheck(BaseMeasurement):
    test_keys: DNSCheckTestKeys


@dataclass
class TorTestTarget(BaseModel):
    target_address: str
    target_protocol: str

    network_events: Optional[List[NetworkEvent]] = None
    queries: Optional[List[DNSQuery]] = None
    requests: Optional[List[HTTPTransaction]] = None
    tls_handshakes: Optional[List[TLSHandshake]] = None
    tcp_connect: Optional[List[TCPConnect]] = None
    target_name: Optional[str] = None
    failure: Failure = None


@dataclass
class TorTestKeys(BaseModel):
    targets: Dict[str, TorTestTarget]


@dataclass
class Tor(BaseMeasurement):
    test_keys: TorTestKeys


nettest_dataformats = {
    "web_connectivity": WebConnectivity,
    "tor": Tor,
    "dnscheck": DNSCheck,
    "whatsapp": Whatsapp,
    "signal": Signal,
}

SupportedDataformats = Union[WebConnectivity, Tor, DNSCheck, Whatsapp, BaseMeasurement]


def load_measurement(
    msmt: Optional[dict] = None, msmt_path: Optional[Path] = None
) -> SupportedDataformats:
    if msmt_path:
        with msmt_path.open() as in_file:
            msmt = orjson.loads(in_file.read())

    assert msmt, "either msmt or msmt_path should be set"
    dc = nettest_dataformats.get(msmt["test_name"], BaseMeasurement)
    return dc.from_dict(msmt)



from datetime import datetime

from typing import Optional, Tuple, List, Any, Type, Mapping, Dict
from dataclasses import fields
from oonidata.observations import (
    NettestObservation,
    Observation,
    DNSObservation,
    TCPObservation,
    TLSObservation,
    HTTPObservation,
)
from oonidata.experiments.experiment_result import (
    BlockingType,
    ExperimentResult,
    BlockingEvent,
)


def typing_to_clickhouse(t: Any) -> str:
    if t == str:
        return "String"

    if t in (Optional[str], Optional[bytes]):
        return "Nullable(String)"

    if t == int:
        return "Int32"

    if t == Optional[int]:
        return "Nullable(Int32)"

    if t == bool:
        return "Int8"

    if t == Optional[bool]:
        return "Nullable(Int8)"

    if t == datetime:
        return "Datetime64(6)"

    if t == Optional[datetime]:
        return "Nullable(Datetime64(6))"

    if t == float:
        return "Float64"

    if t == dict:
        return "JSON"

    if t == Optional[float]:
        return "Nullable(Float64)"

    if t in (List[str], List[bytes]):
        return "Array(String)"

    if t == Optional[List[str]]:
        return "Nullable(Array(String))"

    if t == Optional[List[Tuple[str, bytes]]]:
        return "Nullable(Array(Array(String)))"

    if t == BlockingType:
        return "String"

    if t == List[BlockingEvent]:
        columns = []
        for name, type in BlockingEvent.__annotations__.items():
            type_str = typing_to_clickhouse(type)
            columns.append(f"         {name} {type_str}")
        columns_str = ",\n".join(columns)

        s = "Nested (\n"
        s += columns_str
        s += "\n     )"
        return s

    if t in (Mapping[str, str], Dict[str, str]):
        return "Map(String, String)"

    raise Exception(f"Unhandled type {t}")


def create_query_for_observation(obs_class: Type[Observation]) -> Tuple[str, str]:
    columns = []
    for f in fields(obs_class):
        type_str = typing_to_clickhouse(f.type)
        columns.append(f"     {f.name} {type_str}")

    columns_str = ",\n".join(columns)

    return (
        f"""
    CREATE TABLE {obs_class.__table_name__} (
{columns_str}
    )
    ENGINE = ReplacingMergeTree
    ORDER BY (timestamp, observation_id, measurement_uid)
    SETTINGS index_granularity = 8192;
    """,
        obs_class.__table_name__,
    )


def create_query_for_experiment_result() -> Tuple[str, str]:
    columns = []
    for f in fields(ExperimentResult):
        type_str = typing_to_clickhouse(f.type)
        columns.append(f"     {f.name} {type_str}")

    columns_str = ",\n".join(columns)

    return (
        f"""
    CREATE TABLE experiment_result (
{columns_str}
    )
    ENGINE = ReplacingMergeTree
    ORDER BY (timestamp, observation_id, measurement_uid)
    SETTINGS index_granularity = 8192;
    """,
        "experiment_result",
    )


def main():
    create_queries = [
        create_query_for_observation(DNSObservation),
        create_query_for_observation(TCPObservation),
        create_query_for_observation(TLSObservation),
        create_query_for_observation(HTTPObservation),
        create_query_for_observation(NettestObservation),
        create_query_for_experiment_result(),
        (
            """
        CREATE TABLE dns_consistency_tls_baseline (
            ip String,
            domain_name String,
            timestamp Datetime
        )
        ENGINE = ReplacingMergeTree
        ORDER BY (ip, domain_name, timestamp)
        SETTINGS index_granularity = 8192;
        """,
            "dns_consistency_tls_baseline",
        ),
    ]
    for query, table_name in create_queries:
        print(f"clickhouse-client -q 'DROP TABLE {table_name}';")
        print("cat <<EOF | clickhouse-client -nm")
        print(query)
        print("EOF")


if __name__ == "__main__":
    main()



import sys
import gzip
import logging
import multiprocessing
from pathlib import Path
from typing import List, Optional
from datetime import date, timedelta, datetime
from typing import List, Optional

import click

from oonidata import __version__
from oonidata.dataclient import (
    date_interval,
    sync_measurements,
)
from oonidata.db.connections import CSVConnection, ClickhouseConnection
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB
from oonidata.processing import process_day


log = logging.getLogger("oonidata")

log.addHandler(logging.StreamHandler())
log.setLevel(logging.INFO)


def _parse_date(ctx, param, date_str: str) -> date:
    return datetime.strptime(date_str, "%Y-%m-%d").date()


def _parse_csv(ctx, param, s: Optional[str]) -> List[str]:
    if s:
        return s.split(",")
    return []


probe_cc_option = click.option(
    "--probe-cc",
    callback=_parse_csv,
    help="two letter country code, can be comma separated for a list (eg. IT,US). If omitted will select process all countries.",
)
test_name_option = click.option(
    "--test-name",
    type=str,
    callback=_parse_csv,
    help="test_name you care to process, can be comma separated for a list (eg. web_connectivity,whatsapp). If omitted will select process all test names.",
)
start_day_option = click.option(
    "--start-day",
    default=date.today() - timedelta(days=14),
    callback=_parse_date,
    help="""the timestamp of the day for which we should start processing data (inclusive).

    Note: this is the upload date, which doesn't necessarily match the measurement date.
    """,
)
end_day_option = click.option(
    "--end-day",
    default=date.today() + timedelta(days=1),
    callback=_parse_date,
    help="""the timestamp of the day for which we should start processing data (inclusive). 

    Note: this is the upload date, which doesn't necessarily match the measurement date.
    """,
)


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.option("--output-dir", type=Path, required=True)
@probe_cc_option
@test_name_option
@start_day_option
@end_day_option
@click.option("--max-string-size", type=int)
def sync(
    output_dir: Path,
    probe_cc: List[str],
    test_name: List[str],
    start_day: date,
    end_day: date,
    max_string_size: Optional[int] = None,
):
    """
    Sync OONI measurements to a directory
    """
    click.echo(
        f"Downloading measurements for {start_day} - {end_day} into {output_dir}"
    )
    if probe_cc:
        click.echo(f"probe_cc: {','.join(probe_cc)}")
    if test_name:
        click.echo(f"test_name: {','.join(test_name)}")

    sync_measurements(
        output_dir=output_dir,
        probe_cc=probe_cc,
        test_name=test_name,
        start_day=start_day,
        end_day=end_day,
        max_string_size=max_string_size,
    )


def processing_worker(
    day_queue: multiprocessing.Queue,
    data_dir: Path,
    probe_cc: List[str],
    test_name: List[str],
    clickhouse: Optional[str],
    csv_dir: Optional[str],
    start_at_idx: int,
    fast_fail: bool,
):
    fingerprintdb = FingerprintDB(datadir=data_dir, download=False)
    netinfodb = NetinfoDB(datadir=data_dir, download=False)

    if clickhouse:
        db = ClickhouseConnection(clickhouse)
    elif csv_dir:
        db = CSVConnection(csv_dir)
    else:
        raise Exception("Missing --csv-dir or --clickhouse")

    while True:
        day = day_queue.get(block=True)
        if day == None:
            break
        process_day(
            db,
            fingerprintdb,
            netinfodb,
            day,
            test_name=test_name,
            probe_cc=probe_cc,
            start_at_idx=start_at_idx,
            fast_fail=fast_fail,
        )

    db.close()


@cli.command()
@probe_cc_option
@test_name_option
@start_day_option
@end_day_option
@click.option("--csv-dir", type=Path)
@click.option("--clickhouse", type=str)
@click.option(
    "--data-dir",
    type=Path,
    required=True,
    help="data directory to store fingerprint and geoip databases",
)
@click.option(
    "--parallelism",
    type=int,
    help="number of processes to use. Only works when writing to a database",
)
@click.option("--start-at-idx", type=int, default=0)
@click.option("--fast-fail", default=False)
def mkobs(
    probe_cc: List[str],
    test_name: List[str],
    start_day: date,
    end_day: date,
    csv_dir: Optional[Path],
    clickhouse: Optional[str],
    data_dir: Path,
    parallelism: int,
    start_at_idx: int,
    fast_fail: bool,
):
    """
    Make observations for OONI measurements and write them into clickhouse or a CSV file
    """
    FingerprintDB(datadir=data_dir, download=True)
    NetinfoDB(datadir=data_dir, download=True)

    if csv_dir:
        click.echo(
            "When generating CSV outputs we currently only support parallelism of 1"
        )
        parallelism = 1

    day_queue = multiprocessing.Queue()
    pool = multiprocessing.Pool(
        processes=parallelism,
        initializer=processing_worker,
        initargs=(
            day_queue,
            data_dir,
            probe_cc,
            test_name,
            clickhouse,
            csv_dir,
            start_at_idx,
            fast_fail,
        ),
    )
    for day in date_interval(start_day, end_day):
        day_queue.put(day)

    for _ in range(parallelism):
        day_queue.put(None)

    day_queue.close()
    day_queue.join_thread()

    pool.close()
    pool.join()


if __name__ == "__main__":
    cli()



import re
import csv

from pathlib import Path
from collections import OrderedDict
from typing import Optional, List, Dict
from dataclasses import dataclass, field

import requests

from oonidata.dataformat import HTTPResponse
from oonidata.datautils import get_first_http_header_str


@dataclass
class Fingerprint:
    name: str
    pattern: str
    pattern_type: str
    location_found: str
    other_names: str
    exp_url: Optional[str] = ""
    confidence_no_fp: Optional[int] = 5
    source: List[Optional[str]] = field(default_factory=list)
    scope: Optional[str] = ""
    notes: Optional[str] = ""
    expected_countries: Optional[List[str]] = field(default_factory=list)

    regexp: Optional[re.Pattern] = None

    def matches_pattern(self, s: str) -> bool:
        if self.pattern_type == "full":
            return self.pattern == s

        if self.pattern_type == "contains":
            return self.pattern in s

        if self.pattern_type == "prefix":
            return s.startswith(self.pattern)

        if self.pattern_type == "regexp":
            assert (
                self.regexp is not None
            ), "regexp is not set for a regexp type pattern"
            return self.regexp.search(s) != None

        raise Exception(
            f"Found unknown fingerprint matching pattern {self.pattern_type}"
        )

    def matches_http(self, http_response: HTTPResponse) -> bool:
        assert self.location_found != "dns", "Cannot use a DNS signature to match HTTP"

        if self.location_found == "body":
            if not http_response.body_str:
                return False

            try:
                return self.matches_pattern(http_response.body_str)
            except UnicodeDecodeError:
                return False

        if self.location_found.startswith("header.") and self.regexp:
            search_header = self.location_found[len("header.") :]
            header_value = get_first_http_header_str(
                search_header, http_response.headers_list_str or []
            )
            if not header_value:
                return False

            assert isinstance(header_value, str)

            return self.matches_pattern(header_value)
        return False


def _load_fingerprints(filepath: Path) -> Dict[str, Fingerprint]:
    fingerprints = OrderedDict()

    with filepath.open() as in_file:
        csv_reader = csv.DictReader(in_file)
        for row in csv_reader:
            fp = Fingerprint(**row)
            if fp.pattern_type == "regexp":
                fp.regexp = re.compile(fp.pattern, re.DOTALL)
            fingerprints[fp.name] = fp
    return fingerprints


class FingerprintDB:
    def __init__(self, datadir: Path, download: bool = False):
        self.datadir = datadir
        self.fingerprint_dir = datadir / "blocking-fingerprints"

        dns_fp_path = self.fingerprint_dir / "fingerprints_dns.csv"
        http_fp_path = self.fingerprint_dir / "fingerprints_http.csv"

        if download and (not dns_fp_path.exists() or not http_fp_path.exists()):
            self.refresh_fingerprintdb()

        self.dns_fp = _load_fingerprints(dns_fp_path)
        self.http_fp = _load_fingerprints(http_fp_path)

    def refresh_fingerprintdb(self):
        self.fingerprint_dir.mkdir(parents=True, exist_ok=True)

        for fn in ["fingerprints_http.csv", "fingerprints_dns.csv"]:
            with requests.get(
                f"https://raw.githubusercontent.com/ooni/blocking-fingerprints/main/{fn}",
                stream=True,
            ) as resp:
                resp.raise_for_status()
                output_path = self.fingerprint_dir / fn
                with output_path.with_suffix(".tmp").open("wb") as out_file:
                    for b in resp.iter_content(chunk_size=2**16):
                        out_file.write(b)

            output_path.with_suffix(".tmp").rename(output_path)

    def match_http(self, http_response: HTTPResponse) -> List[Fingerprint]:
        matches = []
        for fp in self.http_fp.values():
            if fp.matches_http(http_response):
                matches.append(fp)
        return matches

    def match_dns(self, address: Optional[str]) -> Optional[Fingerprint]:
        if not address:
            return None
        for fp in self.dns_fp.values():
            if fp.pattern == address:
                return fp
        return None

    def get_fp(self, name: str) -> Fingerprint:
        try:
            return self.dns_fp[name]
        except KeyError:
            return self.http_fp[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oonidata', 'oonidata.cli', 'oonidata.db', 'oonidata.experiments']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0',
 'boto3>=1.24',
 'click>=8.0.0',
 'clickhouse-driver>=0.2',
 'cryptography>=38.0.3',
 'lxml>=4.9',
 'lz4>=4.0',
 'mashumaro>=3.0',
 'maxminddb>=2.2',
 'orjson>=3.8',
 'pyOpenSSL>=22.1',
 'requests>=2.27',
 'tqdm>=4.64']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0']}

entry_points = \
{'console_scripts': ['oonidata = oonidata.cli:cli']}

setup_kwargs = {
    'name': 'oonidata',
    'version': '0.2.3',
    'description': '',
    'long_description': None,
    'author': 'Arturo Filastò',
    'author_email': 'arturo@filasto.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



import csv

from collections import namedtuple
from datetime import datetime
from pprint import pformat
import logging

log = logging.getLogger("oonidata.processing")


class DatabaseConnection:
    def __init__(self):
        self.client = None

    def write_row(self, table_name, row):
        log.info(f"Writing to {table_name}")
        log.info(pformat(row))

    def close(self):
        pass


class ClickhouseConnection(DatabaseConnection):
    def __init__(self, conn_url):
        from clickhouse_driver import Client

        self.client = Client.from_url(conn_url)

    def execute(self, *args, **kwargs):
        # log.debug(f"execute {args} {kwargs}")
        return self.client.execute(*args, **kwargs)

    def write_row(self, table_name, row):
        fields = ", ".join(row.keys())
        query_str = f"INSERT INTO {table_name} ({fields}) VALUES"
        try:
            self.client.execute(query_str, [row])
        except Exception as exc:
            log.error(f"Failed to write row")
            log.error(pformat(row))
            raise exc


CSVConnectionHandle = namedtuple("CSVConnectionHandle", ["fh", "writer"])


class CSVConnection(DatabaseConnection):
    def __init__(self, output_dir):
        self.output_dir = output_dir
        self.open_writers = {}

    def write_row(self, table_name, row):
        if table_name not in self.open_writers:
            ts = datetime.utcnow().strftime("%Y%m%dT%H%M%S")
            fh = (self.output_dir / f"{table_name}-{ts}.csv").open("w")
            csv_writer = csv.DictWriter(fh, fieldnames=list(row.keys()))
            csv_writer.writeheader()
            self.open_writers[table_name] = CSVConnectionHandle(
                writer=csv_writer, fh=fh
            )

        self.open_writers[table_name].writer.writerow(row)

    def close(self):
        for handle in self.open_writers.values():
            handle.fh.close()



from dataclasses import dataclass, field
from datetime import date
import logging

from typing import Any, Dict, Optional, Tuple, List
from urllib.parse import urlparse
from oonidata.dataformat import WebConnectivity, WebConnectivityControl
from oonidata.datautils import one_day_dict

from oonidata.db.connections import ClickhouseConnection

log = logging.getLogger("oonidata.processing")


@dataclass
class TCPControl:
    address: str
    reachable_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    unreachable_cc_asn: List[Tuple[str, int]] = field(default_factory=list)


def make_tcp_control_map(
    day: date, domain_name: str, db: ClickhouseConnection
) -> Dict[str, TCPControl]:
    tcp_control_map = {}
    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT probe_cc, probe_asn, ip, port, failure FROM obs_tcp
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, ip, port, failure;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, ip, port, failure in res:
            address = f"{ip}:{port}"
            tcp_control_map[address] = tcp_control_map.get(address, TCPControl(address))
            if not failure:
                tcp_control_map[address].reachable_cc_asn.append((probe_cc, probe_asn))
            else:
                tcp_control_map[address].unreachable_cc_asn.append(
                    (probe_cc, probe_asn)
                )
    return tcp_control_map


@dataclass
class HTTPControl:
    url: str
    failure_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    ok_cc_asn: List[Tuple[str, int]] = field(default_factory=list)

    response_body_length: int = 0
    response_body_sha1: str = ""
    response_body_title: str = ""
    response_body_meta_title: str = ""

    response_status_code: int = 0


def maybe_get_first(l: list, default_value: Any = None) -> Optional[Any]:
    try:
        return l[0]
    except IndexError:
        return default_value


def make_http_control_map(
    day: date, domain_name: str, db: ClickhouseConnection
) -> Dict[str, HTTPControl]:
    http_control_map = {}

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT probe_cc, probe_asn, request_url, failure FROM obs_http
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, request_url, failure;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, request_url, failure in res:
            http_control_map[request_url] = http_control_map.get(
                request_url, HTTPControl(request_url)
            )
            if not failure:
                http_control_map[request_url].ok_cc_asn.append((probe_cc, probe_asn))
            else:
                http_control_map[request_url].failure_cc_asn.append(
                    (probe_cc, probe_asn)
                )

    q = """SELECT request_url,
    topK(1)(response_body_sha1),
    topK(1)(response_body_length),
    topK(1)(response_body_title),
    topK(1)(response_body_meta_title),
    topK(1)(response_status_code)
    FROM obs_http
    WHERE failure IS NULL
    AND domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY request_url;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for (
            request_url,
            response_body_sha1,
            response_body_length,
            response_body_title,
            response_body_meta_title,
            response_status_code,
        ) in res:
            http_control_map[request_url] = http_control_map.get(
                request_url, HTTPControl(request_url)
            )
            http_control_map[request_url].response_body_sha1 = maybe_get_first(
                response_body_sha1, ""
            )
            http_control_map[request_url].response_body_length = maybe_get_first(
                response_body_length, ""
            )
            http_control_map[request_url].response_body_title = maybe_get_first(
                response_body_title, ""
            )
            http_control_map[request_url].response_body_meta_title = maybe_get_first(
                response_body_meta_title, ""
            )
            http_control_map[request_url].response_status_code = maybe_get_first(
                response_status_code, ""
            )

    return http_control_map


@dataclass
class DNSControl:
    domain: str
    nxdomain_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    failure_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    ok_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    tls_consistent_answers: List[str] = field(default_factory=list)
    answers_map: Dict[str, List[Tuple[str, str]]] = field(default_factory=dict)


def make_dns_control(
    day: date, domain_name: str, db: ClickhouseConnection
) -> DNSControl:
    dns_baseline = DNSControl(domain_name)

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT DISTINCT(ip) FROM obs_tls
    WHERE is_certificate_valid = 1 
    AND domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        dns_baseline.tls_consistent_answers = [row[0] for row in res]

    q = """SELECT probe_cc, probe_asn, failure, answer FROM obs_dns
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, failure, answer;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, failure, ip in res:
            if not failure:
                dns_baseline.ok_cc_asn.append((probe_cc, probe_asn))
                dns_baseline.answers_map[probe_cc] = dns_baseline.answers_map.get(
                    probe_cc, []
                )
                if ip:
                    dns_baseline.answers_map[probe_cc].append((probe_asn, ip))
                else:
                    log.error(
                        f"No IP present for {domain_name} {probe_cc} ({probe_asn}) in baseline"
                    )
            else:
                dns_baseline.failure_cc_asn.append((probe_cc, probe_asn))
                if failure == "dns_nxdomain_error":
                    dns_baseline.nxdomain_cc_asn.append((probe_cc, probe_asn))

    return dns_baseline


def make_dns_control_from_wc(
    msmt_input: str, control: WebConnectivityControl
) -> DNSControl:
    domain_name = urlparse(msmt_input).hostname

    assert domain_name is not None, "domain_name is None"

    if not control or not control.dns:
        return DNSControl(domain=domain_name)

    nxdomain_cc_asn = []
    if control.dns.failure == "dns_nxdomain_error":
        nxdomain_cc_asn.append(("ZZ", 0))

    ok_cc_asn = []
    failure_cc_asn = []
    if control.dns.failure is not None:
        failure_cc_asn.append(("ZZ", 0))
    else:
        ok_cc_asn.append(("ZZ", 0))

    answers_map = {}
    if control.dns.addrs:
        answers_map["ZZ"] = [(0, ip) for ip in control.dns.addrs]

    return DNSControl(
        domain=domain_name,
        answers_map=answers_map,
        ok_cc_asn=ok_cc_asn,
        nxdomain_cc_asn=nxdomain_cc_asn,
        failure_cc_asn=failure_cc_asn,
    )


def make_tcp_control_from_wc(
    control: WebConnectivityControl,
) -> Dict[str, TCPControl]:
    if not control or not control.tcp_connect:
        return {}

    tcp_b_map = {}
    for key, status in control.tcp_connect.items():
        if status.failure == None:
            tcp_b_map[key] = TCPControl(address=key, reachable_cc_asn=[("ZZ", 0)])
        else:
            tcp_b_map[key] = TCPControl(address=key, unreachable_cc_asn=[("ZZ", 0)])
    return tcp_b_map


def make_http_control_from_wc(
    msmt: WebConnectivity, control: WebConnectivityControl
) -> Dict[str, HTTPControl]:
    if not control or not control.http_request:
        return {}

    if not msmt.test_keys.requests:
        return {}

    http_b_map = {}
    # We make the baseline apply to every URL in the response chain, XXX evaluate how much this is a good idea
    for http_transaction in msmt.test_keys.requests:
        if not http_transaction.request:
            continue

        url = http_transaction.request.url
        if control.http_request.failure == None:
            http_b_map[url] = HTTPControl(
                url=url,
                response_body_title=control.http_request.title or "",
                response_body_length=control.http_request.body_length or 0,
                response_status_code=control.http_request.status_code or 0,
                response_body_meta_title="",
                response_body_sha1="",
                ok_cc_asn=[("ZZ", 0)],
            )
        else:
            http_b_map[url] = HTTPControl(
                url=url,
                response_body_title="",
                response_body_length=0,
                response_status_code=0,
                response_body_meta_title="",
                response_body_sha1="",
                failure_cc_asn=[("ZZ", 0)],
            )
    return http_b_map



from typing import List
from oonidata.dataformat import SIGNAL_PEM_STORE
from oonidata.datautils import InvalidCertificateChain, TLSCertStore
from oonidata.experiments.experiment_result import (
    BlockingEvent,
    BlockingType,
    ExperimentResult,
    fp_scope_to_outcome,
    make_base_result_meta,
)
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB
from oonidata.observations import (
    DNSObservation,
    HTTPObservation,
    NettestObservation,
    TCPObservation,
    TLSObservation,
)


class SignalExperimentResult(ExperimentResult):
    pass


def make_signal_experiment_result(
    nt_obs: NettestObservation,
    dns_o_list: List[DNSObservation],
    tcp_o_list: List[TCPObservation],
    tls_o_list: List[TLSObservation],
    http_o_list: List[HTTPObservation],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> SignalExperimentResult:
    # We got nothin', we can't do nothin'
    assert len(dns_o_list) > 0, "empty DNS observation list"

    confirmed = False
    anomaly = False

    signal_is_down = False

    blocking_events = []
    observation_ids = []
    inconsistent_dns_answers = set()

    for dns_o in dns_o_list:
        observation_ids.append(dns_o.observation_id)
        if dns_o.domain_name == "uptime.signal.org":
            # This DNS query is used by signal to figure out if some of it's
            # services are down.
            # see: https://github.com/signalapp/Signal-Android/blob/c4bc2162f23e0fd6bc25941af8fb7454d91a4a35/app/src/main/java/org/thoughtcrime/securesms/jobs/ServiceOutageDetectionJob.java#L25
            # TODO: should we do something in the case in which we can't tell
            # because DNS blocking is going on (ex. in Iran)?
            if dns_o.answer == "127.0.0.2":
                signal_is_down = True
            continue

        if dns_o.failure:
            anomaly = True
            blocking_events.append(
                BlockingEvent(
                    blocking_type=BlockingType.BLOCKED,
                    blocking_subject=f"{dns_o.domain_name}",
                    blocking_detail=f"dns.{dns_o.failure}",
                    blocking_meta={},
                    confidence=0.8,
                )
            )
            continue

        if not dns_o.is_tls_consistent:
            # We don't set the anomaly flag, because this logic is very
            # susceptible to false positives
            # anomaly = True
            blocking_meta = {"why": "tls_inconsistent", "ip": dns_o.answer}
            blocking_type = BlockingType.BLOCKED
            fp = fingerprintdb.match_dns(dns_o.answer)
            if fp:
                blocking_type = fp_scope_to_outcome(fp.scope)
                if blocking_type != BlockingType.SERVER_SIDE_BLOCK:
                    confirmed = True
                    anomaly = True
                blocking_meta["fp_name"] = fp.name

            confidence = 0.3
            # Having a TLS inconsistency is a much stronger indication than not
            # knowing.
            if dns_o.is_tls_consistent == False:
                inconsistent_dns_answers.add(dns_o.answer)
                # Only set the anomaly here
                anomaly = True
                confidence = 0.8

            blocking_events.append(
                BlockingEvent(
                    blocking_type=blocking_type,
                    blocking_detail="dns.inconsistent",
                    blocking_subject=f"{dns_o.domain_name}",
                    blocking_meta=blocking_meta,
                    confidence=confidence,
                )
            )
            continue

    if signal_is_down:
        # The service is down. No point in going on with the analysis
        # It's still possible for the service to be down, yet there to be DNS
        # level interference, so we still count the other blocking_events if they were
        # present.
        blocking_events.append(
            BlockingEvent(
                blocking_type=BlockingType.DOWN,
                blocking_subject="Signal Messenger",
                blocking_detail="down",
                blocking_meta={},
                confidence=0.9,
            )
        )
        return SignalExperimentResult(
            blocking_events=blocking_events,
            observation_ids=observation_ids,
            anomaly=anomaly,
            confirmed=confirmed,
            **make_base_result_meta(nt_obs),
        )

    for tcp_o in tcp_o_list:
        observation_ids.append(tcp_o.observation_id)
        if tcp_o.failure and tcp_o.ip not in inconsistent_dns_answers:
            anomaly = True
            blocking_events.append(
                BlockingEvent(
                    blocking_type=BlockingType.BLOCKED,
                    blocking_subject=f"{tcp_o.ip}:{tcp_o.port}",
                    blocking_detail=f"tcp.{tcp_o.failure}",
                    blocking_meta={},
                    confidence=0.7,
                )
            )

    cert_store = TLSCertStore(pem_cert_store=SIGNAL_PEM_STORE)
    for tls_o in tls_o_list:
        observation_ids.append(tls_o.observation_id)
        # We skip analyzing TLS handshakes that are the result of an
        # inconsistent DNS resolution.
        if tls_o.ip in inconsistent_dns_answers:
            continue

        if tls_o.failure and not tls_o.failure.startswith("ssl_"):
            anomaly = True
            blocking_events.append(
                BlockingEvent(
                    blocking_type=BlockingType.BLOCKED,
                    blocking_subject=f"{tls_o.server_name}",
                    blocking_detail=f"tls.{tls_o.failure}",
                    blocking_meta={},
                    confidence=0.7,
                )
            )

        if tls_o.peer_certificates:
            try:
                _, _ = cert_store.validate_cert_chain(
                    tls_o.timestamp, tls_o.peer_certificates
                )
                # The server_name is listed in the SAN only for the older certs.
                # Since we are pinning to only the two known signal CAs it's
                # probably safe to just ignore.
            except InvalidCertificateChain as exc:
                anomaly = True
                blocking_events.append(
                    BlockingEvent(
                        blocking_type=BlockingType.BLOCKED,
                        blocking_subject=f"{tls_o.server_name}",
                        blocking_detail=f"tls.ssl_invalid_certificate",
                        blocking_meta={"cert_error": str(exc)},
                        confidence=0.9,
                    )
                )

    # This is an upper bound, which means we might be over-estimating blocking
    ok_confidence = 1 - max(map(lambda o: o.confidence, blocking_events), default=0)
    blocking_events.append(
        BlockingEvent(
            blocking_type=BlockingType.OK,
            blocking_subject=f"Signal Messenger",
            blocking_detail=f"ok",
            blocking_meta={},
            confidence=ok_confidence,
        )
    )

    # TODO: add support for validating if the HTTP responses are also consistent
    return SignalExperimentResult(
        blocking_events=blocking_events,
        observation_ids=observation_ids,
        anomaly=anomaly,
        confirmed=confirmed,
        ok_confidence=ok_confidence,
        **make_base_result_meta(nt_obs),
    )
